#![allow(unused_imports)]
use super::*;
use wasm_bindgen::prelude::*;
#[wasm_bindgen]
extern "C" {
    # [wasm_bindgen (extends = SvgElement , extends = Element , extends = Node , extends = EventTarget , extends = :: js_sys :: Object , js_name = SVGFEFloodElement , typescript_type = "SVGFEFloodElement")]
    #[derive(Debug, Clone, PartialEq, Eq)]
    #[doc = "The `SvgfeFloodElement` class."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgfeFloodElement`*"]
    pub type SvgfeFloodElement;
    #[cfg(feature = "SvgAnimatedLength")]
    # [wasm_bindgen (structural , method , getter , js_class = "SVGFEFloodElement" , js_name = x)]
    #[doc = "Getter for the `x` field of this object."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement/x)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgAnimatedLength`, `SvgfeFloodElement`*"]
    pub fn x(this: &SvgfeFloodElement) -> SvgAnimatedLength;
    #[cfg(feature = "SvgAnimatedLength")]
    # [wasm_bindgen (structural , method , getter , js_class = "SVGFEFloodElement" , js_name = y)]
    #[doc = "Getter for the `y` field of this object."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement/y)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgAnimatedLength`, `SvgfeFloodElement`*"]
    pub fn y(this: &SvgfeFloodElement) -> SvgAnimatedLength;
    #[cfg(feature = "SvgAnimatedLength")]
    # [wasm_bindgen (structural , method , getter , js_class = "SVGFEFloodElement" , js_name = width)]
    #[doc = "Getter for the `width` field of this object."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement/width)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgAnimatedLength`, `SvgfeFloodElement`*"]
    pub fn width(this: &SvgfeFloodElement) -> SvgAnimatedLength;
    #[cfg(feature = "SvgAnimatedLength")]
    # [wasm_bindgen (structural , method , getter , js_class = "SVGFEFloodElement" , js_name = height)]
    #[doc = "Getter for the `height` field of this object."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement/height)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgAnimatedLength`, `SvgfeFloodElement`*"]
    pub fn height(this: &SvgfeFloodElement) -> SvgAnimatedLength;
    #[cfg(feature = "SvgAnimatedString")]
    # [wasm_bindgen (structural , method , getter , js_class = "SVGFEFloodElement" , js_name = result)]
    #[doc = "Getter for the `result` field of this object."]
    #[doc = ""]
    #[doc = "[MDN Documentation](https://developer.mozilla.org/en-US/docs/Web/API/SVGFEFloodElement/result)"]
    #[doc = ""]
    #[doc = "*This API requires the following crate features to be activated: `SvgAnimatedString`, `SvgfeFloodElement`*"]
    pub fn result(this: &SvgfeFloodElement) -> SvgAnimatedString;
}
