use lopdf::{Document, Object, Stream};

fn main() {
    let doc = Document::load("/Users/Junfeng/Downloads/四年级/asm.pdf").unwrap();
    println!("version: {:#?}, max_id: {:#?}", &doc.version, &doc.max_id);
    let pages = doc.get_pages().iter()
        .map(|page| println!("pagenum: {:#?}, pageid: {:#?}",
                          page.0, page.1)).count();
    println!("{}", pages);
}
